#undef LOCAL_DEBUG

#include "stroj_casu.h"

#include <termios.h>
#include <ucw/chartype.h>

/*** Notifiers ***/

// Boolean
int state_left_sw, state_right_sw, state_black_button, state_red_button,
    state_lock, state_bell_button, state_magnet, state_big_red_sw, state_steam,
    state_antenna;

// Analog (0-255)
int state_tap, state_solar;

// Set if notify is caused by a reboot notification instead of a normal status report packet
int hardware_rebooted;

void notify_left_sw(void)
{
	if (hardware_rebooted) return;
	lcd_printf("ON -> ON\n");
}

void notify_right_sw(void)
{
	if (hardware_rebooted) return;
	lcd_printf("OFF -> OFF\n");
}

void notify_black_button(void)
{
	if (state_black_button) return;
	if (hardware_rebooted) return;
	if (noplease())
		lcd_printf("Pluvia et sol.\n");
	msg(L_DEBUG, "HW: black button\n");
//	weather(stroj_now);
}

void notify_red_button(void)
{
	if (state_red_button) return;
	if (hardware_rebooted) return;
	msg(L_DEBUG, "HW: red button\n");
	if (please())
		fortune(NULL);
}

void notify_lock(void)
{
/*	if (state_lock)
		GP_mode_init();
	else
		GP_mode_stop();*/
}

int ___diode_on_off_begin;
int ___diode_on_off_reading;
#define	___DIODE_ON_OFF_BEGIN	8270

int gp_mode;
int gp_mode_number;

void gp_mode_confirm(void)
{
	struct years {
		int real;
		int im;
	} year;
	struct years possible_years[] = {
 		{-334,-1},
		{1866,0},
 		{0,15},
		{1770,0},
 		{1933,6},
		
		{1360,0},
 		{1227,0},
		{0,-2000},
 		{1960,-0},
		{1592,0},
		
 		{1624,0},
		{2100,10},
 		{0,-666},
		{-1000,-200},
 		{1886,0},
		
		{1952,3},
 		{1960,0},
		{1930,0},
 		{-350,-2},
		{1704,4},
		
 		{1350,60},
		{2020,0},
 		{-250,10},
		{1680,0},
 		{1930,0},
		
		{1619,0},
 		{-850,1},
		{1321,50},
 		{1800,1000},
		{2009, 0},

 		{3325, 0}
			};
	int imagpart, realpart;
	int targetyear;
	if (gp_mode == 1)
	{
		gp_mode_number = -1;
		printf("Entering code ...\n");
		gp_mode = 2;
	}
	else
	{
		printf("Code entered.\n");
		if (gp_mode_number < 10000) 
		{
		 	printf("Not long enough.\n");
			goto end;
		}
		gp_mode = 1;
		if (gp_mode_number % 23 == 0)
		{
			lcd_printf("bad number\n");
			printf("bad number\n");
			goto end;
		}
		targetyear = (gp_mode_number / 17) % 557;
		if ((targetyear % (gp_mode_number % 23)) != 0)
		{
			lcd_printf("bad number\n");
			printf("bad number\n");
			goto end;
		}
		printf("Code entered: %d", gp_mode_number);
		year = possible_years[targetyear % 31];
		imagpart = year.im;
		realpart = year.real;
		lcd_printf("TOKEN OK\n");
		
		if (imagpart == 0)
			lcd_printf("DATE: %d\n", realpart);
		if (imagpart < 0)
			lcd_printf("DATE: %d%di\n", realpart, imagpart);
		if (imagpart > 0)
			lcd_printf("DATE: %d+%di\n", realpart, imagpart);
		gp_mode_number = 0;

	}
end:	switch(gp_mode)
	{
		case 0:	clrled_command("5");
			break;
		case 1: blinkled_command("5");
			break;
		case 2:	setled_command("5");
			break;
	}
}

void notify_bell_button(void)
{
	if (state_bell_button) return;
	if (hardware_rebooted) return;
	if (gp_mode)	gp_mode_confirm();
	if ((main_now_seconds % 86400) < 360*75 || (main_now_seconds % 86400) > 360*225)
		printf("Not ringing. It's night.\n");
	else
		play_sound("ogg/doorbell.ogg");
	
	___diode_on_off_reading = ___diode_on_off_begin;
	___diode_on_off_begin = 0;

}

void notify_magnet(void)
{
	set_green_led(state_magnet);
}

void notify_big_red_sw(void)
{
	if (hardware_rebooted) return;
	if ((random() & 0xFF) > (state_tap / (state_steam ? 1 : 10)))
		play_sound("ogg/steam_out.ogg");
	else
		loc_exec_command(NULL);
/*	if (hardware_rebooted) return;
	msg(L_DEBUG, "Big Red Switch");
	set_next_move(main_now - 365240*(86400)*1700);
	move_in_time();
//	fortune(NULL);*/
}

void notify_tap(void)
{
	if (state_steam)
	{
		set_left_meter(state_tap);
	}
	else
	{
		set_left_meter(state_tap/10);
	}
}

void notify_steam(void)
{
	notify_tap();
}

void notify_antenna(void)
{
}

void notify_solar(void)
{
}

void dialled_digit(uns x)
{
	static u64 began;
	static u64 called;
	static int i;
	if (began < main_now-20000)
	{
		began = main_now;
		called = 0;
		i=0;
	}
	called *= 10;
	called += x;
	printf("Dialled: %Ld\n", called);
	i++;
	if (___diode_on_off_reading == 1)
	{
		switch(x)
		{
			case 1: ___diode_on_off_reading = 3;
				break;
			case 0: ___diode_on_off_reading = 2;
				break;
			default:
				___diode_on_off_reading = 0;
		}
	}
	else if (___diode_on_off_reading)
	{
		switch(x)
		{
			case 1: set_big_red_led(___diode_on_off_reading - 2);
				break;
			case 2: set_big_yellow_led(___diode_on_off_reading - 2);
				break;
			case 3: set_big_green_led(___diode_on_off_reading - 2);
				break;
			case 4: set_lamp(___diode_on_off_reading - 2);
				break;
			case 5: set_orange_led(___diode_on_off_reading - 2);
				break;
			case 6: set_green_led(___diode_on_off_reading - 2);
				break;
			default:
				___diode_on_off_reading = 0;
		}
	}
	if (called == ___DIODE_ON_OFF_BEGIN)
		___diode_on_off_begin = 1;
	if (called % 100 == 42)
		stop_sound();
	if (called % 100000000 == 46864225)
	{
		if (gp_mode == 0)
		{
			gp_mode = 1;
			blinkled_command("5");
			called = 0;
			gp_mode_number = -1;
		}
		else
		{
			gp_mode = 0;
			clrled_command("5");
			called = 0;
		}
	}
	if (i == 10)
	{
		began = 0;
		lcd_printf("%16Ld\n", called);
		play_sound("ogg/dial_rejected.ogg");
		
	}
	struct stat _buf;
	char _com[SHORT_STRING];
	
	sprintf(_com, "ogg/dial/%Ld", called);
	if (gp_mode)
  	{
		if ((gp_mode_number == -1) && (called > 9))
			called %= 10;
		gp_mode_number = called;
	}
	if (stat(_com, &_buf) == -1)
		return;
	
	play_sound(_com);

}

/*** Setters ***/

static void FORMAT_CHECK(printf,1,2) cmd(char *fmt, ...);

// orn --- --- lmp grn GRN YLW RED
static uns last_led_state;

static void send_led_state(void)
{
	cmd("B%02x\n", last_led_state);
}

#define LED(name, mask)							\
	void set_##name(int val)					\
	{								\
		uns new = last_led_state & ~(mask);			\
		if (val) new |= mask;					\
		if (new != last_led_state) {				\
			msg(L_DEBUG, "HW: Setting %s=%d", #name, val);	\
			last_led_state = new;				\
			send_led_state();				\
		}							\
	}

LED(orange_led,		0x80);
LED(lamp,		0x10);
LED(green_led,		0x08);
LED(big_red_led,	0x04);
LED(big_yellow_led,	0x02);
LED(big_green_led,	0x01);

#undef LED

static uns last_adc_l = ~0U, last_adc_r = ~0U;

static void send_adc_state(uns ccmd, uns val)
{
	cmd("%c%02x\n", ccmd, val);
}

void set_left_meter(uns val)
{
	if (last_adc_l != val) {
		last_adc_l = val;
		msg(L_DEBUG, "HW: Setting left meter = %d", val);
		send_adc_state('L', val);
	}
}

void set_right_meter(uns val)
{
	if (last_adc_r != val) {
		last_adc_r = val;
		msg(L_DEBUG, "HW: Setting right meter = %d", val);
		send_adc_state('R', val);
	}
}

/*** Display ***/

#define DISP_LINES 16
#define DISP_DELAY 1000		// ms

static char disp_buffer[DISP_LINES][16];
static int disp_top, disp_w;
static int disp_col;
static timestamp_t disp_last;
static struct main_timer disp_timer;

static void lcd_refresh(void)
{
	DBG("LCD: refresh top=%d w=%d", disp_top, disp_w);
	cmd("G00\n");
	cmd("W%.16s\n", disp_buffer[disp_top]);
	cmd("G10\n");
	cmd("W%.16s\n", disp_buffer[(disp_top+1) % DISP_LINES]);
	disp_last = main_now;
}

static void lcd_delayed(struct main_timer *tm)
{
	disp_top = (disp_top + 1) % DISP_LINES;
	DBG("LCD: delayed top=%d w=%d", disp_top, disp_w);
	lcd_refresh();
	if ((disp_top + 2) % DISP_LINES == disp_w)
		timer_del(tm);
	else
		timer_add(tm, main_now + DISP_DELAY);
}

void lcd_write(char *m)
{
	int c;
	main_get_time();
	while ((c = *m++)) {
		if (c == '\n' || disp_col >= 16) {
			if ((disp_w + 1) % DISP_LINES == disp_top)
				memcpy(disp_buffer[(disp_w - 1 + DISP_LINES) % DISP_LINES] + 11, "(...)", 5);
			else {
				disp_w = (disp_w + 1) % DISP_LINES;
				memset(disp_buffer[disp_w], ' ', 16);
				if (!disp_timer.expires) {
					DBG("LCD: timing %d", (int)(main_now - disp_last));
					if (main_now - disp_last >= DISP_DELAY) {
						disp_top = (disp_top + 1) % DISP_DELAY;
						lcd_refresh();
					} else {
						disp_timer.handler = lcd_delayed;
						timer_add(&disp_timer, disp_last + DISP_DELAY);
					}
				}
			}
			disp_col = 0;
		}
		if (c != '\n') {
			if (c < ' ')
				c = ' ';
			disp_buffer[disp_w][disp_col++] = c;
		}
	}
}

void lcd_printf(char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	char buf[256];
	vsnprintf(buf, sizeof(buf), fmt, args);
	lcd_write(buf);
	va_end(args);
}

static void lcd_init(void)
{
	cmd("C\n");
	memset(disp_buffer[0], ' ', 16);
	memset(disp_buffer[1], ' ', 16);
	memcpy(disp_buffer[1], ";-)", 3);
	memset(disp_buffer[2], ' ', 16);
	disp_w = 2;
	lcd_refresh();
}

/*** Low-level ***/

#define MAX_PKT_SIZE 256

static int serial_fd;

void serial_init(int fd)
{
	struct termios tios = {
		.c_iflag = IGNBRK | IGNPAR | IGNCR,
		.c_oflag = 0,
		.c_cflag = B9600 | CS8 | CREAD | CLOCAL,
		.c_lflag = 0,
		.c_cc[VMIN] = 1,
	};
	if (tcsetattr(fd, TCSANOW, &tios) < 0)
		die("Cannot set termios: %m");
	write(fd, "\r\n", 2);
	serial_fd = fd;
	hardware_rebooted = 1;

	set_left_meter(0);
	set_right_meter(0);
	send_led_state();
	lcd_init();
}

static void cmd(char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	char buf[MAX_PKT_SIZE];
	int l = vsnprintf(buf, sizeof(buf), fmt, args);
	msg(L_DEBUG, "HW: Sending <%s>", buf);
	fcntl(serial_fd, F_SETFL, 0);
	write(serial_fd, buf, l);
	fcntl(serial_fd, F_SETFL, O_NONBLOCK);
	va_end(args);
}

void updatemood(void)
{
	set_right_meter(mood);
}

static int iabs(int x)
{
	return (x < 0) ? -x : x;
}

#define SW(name, from) do {							\
		int new = !!(from);						\
		if (state_##name != new || hardware_rebooted) {			\
			state_##name = new;					\
			msg(L_DEBUG, "HW: Received %s=%d", #name, new);		\
			notify_##name();					\
		}								\
	} while (0)

#define AD(name, new) do {							\
		if (iabs(state_##name - new) > 1 || hardware_rebooted) {			\
			state_##name = new;					\
			msg(L_DEBUG, "HW: Received %s=%d", #name, new);		\
			notify_##name();					\
		}								\
	} while(0)

static void serial_parse_sw(uns a, uns b)
{
	/*
	 *	A	rsw lsw bbt rbt --- --- --- lck
	 *	B	--- --- bel --- mag Rsw --- ---
	 */
	SW(right_sw,		!(a & 0x80));
	SW(left_sw,		!(a & 0x40));
	SW(black_button,	!(a & 0x20));
	SW(red_button,		!(a & 0x10));
	SW(lock,		!(a & 0x01));
	SW(bell_button,		!(b & 0x20));
	SW(magnet,		!(b & 0x08));
	SW(big_red_sw,		!(b & 0x04));
}

static void serial_parse_adc(byte *x)
{
	AD(tap,		x[0]);
	SW(steam,	x[1] >= 4 && x[1] < 16);
	SW(antenna,	x[2] >= 0x20 && x[2] < 0x80);
//	AD(solar,	x[3]);
}

static void serial_parse_dial(uns new)
{
	static uns last_dial;

	if (new != last_dial && (new & 0x0f)) {
		uns d = new & 0x0f;
		if (d == 10)
			d = 0;
		msg(L_DEBUG, "HW: Dialled %d", d);
		dialled_digit(d);
	}
	last_dial = new;
}

#undef SW
#undef AD

static char rxbuf[MAX_PKT_SIZE];
static int rx_size = MAX_PKT_SIZE;

static void serial_parse_pkt(void)
{
	// DBG("HW: [%s]", rxbuf);
	if (!strcmp(rxbuf, "BOOT")) {
		msg(L_INFO, "HW: Rebooted");
		hardware_rebooted = 1;
		send_led_state();
		send_adc_state('L', last_adc_l);
		send_adc_state('R', last_adc_r);
		return;
	}
	if (rxbuf[0] != '<' && rx_size != 15)
		return;

	byte x[7];
	char *p = rxbuf+1;
	for (int i=0; i<7; i++) {
		if (!Cxdigit(p[0]) || !Cxdigit(p[1]))
			msg(L_WARN, "HW: Parse error");
		x[i] = (Cxvalue(p[0]) << 4) | Cxvalue(p[1]);
		p += 2;
	}

	serial_parse_sw(x[0], x[1]);
	serial_parse_adc(x+2);
	serial_parse_dial(x[6]);
	hardware_rebooted = 0;
}

static void serial_parse_char(char c)
{
	if (c == '\r' || c == '\n') {
		if (rx_size < MAX_PKT_SIZE) {
			rxbuf[rx_size] = 0;
			serial_parse_pkt();
		}
		rx_size = 0;
	} else {
		if (rx_size < MAX_PKT_SIZE)
			rxbuf[rx_size++] = c;
	}
}

int serial_parse(struct main_file *fi)
{
	byte buf[256];
	int l = read(fi->fd, buf, sizeof(buf));
	for (int i=0; i<l; i++)
		serial_parse_char(buf[i]);
	return 0;
}

struct hw get_hw_state(void)
{
	struct hw local;
	local.last_led_state = last_led_state;
//	strcpy(&local.lcd_display_text, &lcd_display_text);
	local.last_adc_l = last_adc_l;
	local.last_adc_r = last_adc_r;
	return local;
}

void set_hw_state(struct hw hw)
{
	last_led_state = hw.last_led_state;
	last_adc_l = hw.last_adc_l;
	last_adc_r = hw.last_adc_r;
	send_led_state();
	send_adc_state('L', last_adc_l);
	send_adc_state('R', last_adc_r);
	updatemood();
}

int ___local_leds_blinking;
void blinkled_command(char * wtf)
{
	int i;
	i = atoi(wtf);
	if (i < 1) return;
	if (i > 6) return;
	___local_leds_blinking |= (1 << i);
}

void unblinkled_command(char * wtf)
{
	int i;
	i = atoi(wtf);
	if (i < 1) return;
	___local_leds_blinking &= ~(1 << i);
}

void _blink_on_off(struct main_timer *_mt)
{
	static int last;
	timer_add(_mt, main_now + 500);
	if (___local_leds_blinking & (1 << 1))
		set_big_red_led(last);
	if (___local_leds_blinking & (1 << 2))
		set_big_yellow_led(last);
	if (___local_leds_blinking & (1 << 3))
		set_big_green_led(last);
	if (___local_leds_blinking & (1 << 4))
		set_lamp(last);
	if (___local_leds_blinking & (1 << 5))
		set_orange_led(last);
	if (___local_leds_blinking & (1 << 6))
		set_green_led(last);
	
	last ^= 1;
}

void init_blinking(void)
{
	static struct main_timer _mt;
	_mt.handler = &_blink_on_off;
	timer_add(&_mt, main_now + 500);
}

void setled_command(char * wtf)
{
	switch(atoi(wtf))
	{
		case 1: set_big_red_led(1);
			break;
		case 2: set_big_yellow_led(1);
			break;
		case 3: set_big_green_led(1);
			break;
		case 4: set_lamp(1);
			break;
		case 5: set_orange_led(1);
			break;
		case 6: set_green_led(1);
			break;
	}
	unblinkled_command(wtf);
}

void clrled_command(char * wtf)
{
	switch(atoi(wtf))
	{
		case 1: set_big_red_led(0);
			break;
		case 2: set_big_yellow_led(0);
			break;
		case 3: set_big_green_led(0);
			break;
		case 4: set_lamp(0);
			break;
		case 5: set_orange_led(0);
			break;
		case 6: set_green_led(0);
			break;
	}
	unblinkled_command(wtf);
}
